/* All the functions to deal with the problem of creating a binding
 * for a new packet. */
#include <stddef.h>

#include "mapping-types/ipnat_mapping_generic.c"

/* Mapping types; linked list kept in increasing private_offset order  */
static struct ip_nat_mapping_type *mapping_types;

typedef unsigned long NICE_TYPE;

static const struct ip_nat_rule *
find_rule(const struct ip_nat_pktinfo *pkt)
{
	const struct ip_nat_rule *ptr;

	for (ptr = conns.rule_table; ptr; ptr = ptr->next) {
		struct ip_nat_pktinfo mangled;
		unsigned int i;

		if (ptr->elem.info.hook != pkt->hook)
			continue;

		mangled = *pkt;
#if 1
		DEBUGP("pkt     rule    mask\n");
#endif
		/* Apply mask. */
		for (i = 0; i < sizeof(ptr->elem.mask)/sizeof(NICE_TYPE); i++){
#if 1
			DEBUGP("%08lX %08lX %08lX\n",
			       ((NICE_TYPE *)pkt)[i],
			       ((NICE_TYPE *)&ptr->elem.info)[i],
			       ((NICE_TYPE *)&ptr->elem.mask)[i]);
#endif
			((NICE_TYPE *)&mangled)[i]
				&= ((NICE_TYPE *)&ptr->elem.mask)[i];
		}

		if (memcmp(&ptr->elem.info, &mangled, sizeof(mangled)) == 0)
			break;
	}

	return ptr;
}

static unsigned int ip_nat_do_rule(const struct iphdr *iph,
				   const struct ip_nat_pktinfo *pkt,
				   struct ip_nat_binding **binding)
{
	const struct ip_nat_rule *rule = find_rule(pkt);

	if (!rule) {
		DEBUGP("ip_nat_do_rule: No rule for hook %s.\n",
		       hooknames[pkt->hook]);
		return NF_ACCEPT;
	} else {
		DEBUGP("ip_nat_do_rule: rule for hook %s.\n",
		       hooknames[pkt->hook]);
		return rule->elem.type[0]->rule(iph, pkt, &rule->elem,
						rule->mr, 0, binding);
	}
}

static int mapping_type_cmpfn(const void *arg, const void *curr)
{
	return strcmp(arg, ((struct ip_nat_mapping_type *)curr)->name) == 0;
}

static struct ip_nat_mapping_type *find_mapping_type(const char *name,
						     int load)
{
	struct ip_nat_mapping_type *ret;

	ret = modreg_find(&mapping_types, name, mapping_type_cmpfn);
	if (!ret && load) {
		char modulename[IP_NAT_MAPPING_TYPE_MAX_NAMELEN
			       + sizeof("ip_nat_rule_")];
		strcpy(modulename, "ip_nat_rule_");
		strcat(modulename, name);
		DEBUGP("ip_nat: requesting module `%s'\n", modulename);
		request_module(modulename);
		return find_mapping_type(name, 0);
	}
	else return ret;
}

/* It'd really be easier to simply append to the end, knowing that
 * it'll all come out in the wash.  Of course, combined with module
 * unloading and autoloading, that'd be a *really cool* DoS.  If this
 * weren't free software, trust me, I'd have done it that way.  Maybe
 * there's a moral here... */
static void
find_private_space(struct ip_nat_mapping_type *me)
{
	size_t last_offset = sizeof(struct ip_nat_binding);
	struct ip_nat_mapping_type **ptr;

	/* Go through, add type to wherever there's a big enough hole. */
	down(&modreg_sem);

	for (ptr = &mapping_types; *ptr; ptr = &(*ptr)->next) {
		DEBUGP("find_private_space: %s used %u at offset %u\n",
		       (*ptr)->name, (*ptr)->private_size,
		       (*ptr)->private_offset);
		last_offset = (*ptr)->private_offset + (*ptr)->private_size;
		if ((*ptr)->next
		    && ((*ptr)->next->private_offset - last_offset
			>= me->private_size)) {
			DEBUGP("find_private_space: found %u space\n",
			       me->private_size);
			me->private_offset = last_offset;
			me->next = (*ptr)->next;
			(*ptr)->next = me;
			up(&modreg_sem);
			return;
		}
	}

	/* We're at the end of the list. */
	me->private_offset = last_offset;
	__modreg_append(&mapping_types, me);
	up(&modreg_sem);

	binding_size = last_offset + me->private_size;
	DEBUGP("find_private_space: total size now %u.\n", binding_size);
}

int ip_nat_register_mapping_type(struct ip_nat_mapping_type *me)
{
	if (find_mapping_type(me->name, 0))
		return -EEXIST;

	/* If any fields are NULL, fill them in with noops. */
	if (!me->print_private)
		me->print_private = unknown_proto.mappingtype.print_private;
	if (!me->check)
		me->check = unknown_proto.mappingtype.check;
	if (!me->destroy)
		me->destroy = unknown_proto.mappingtype.destroy;

	MOD_INC_USE_COUNT;
	me->usage = 0;

	find_private_space(me);
	return 0;
}

/* Don't bother reducing binding_size here; it'll be done on next
   find_private_space(). */
int ip_nat_unregister_mapping_type(struct ip_nat_mapping_type *me)
{
	if (me->usage) {
		DEBUGP("ip_nat_unregister_mapping_type: Can't dereg %s\n",
		       me->name);
		return -EBUSY;
	}

	DEBUGP("ip_nat_unregister_mapping_type: dereging %s\n", me->name);
	modreg_delete(&mapping_types, me);
	MOD_DEC_USE_COUNT;
	DEBUGP("ip_nat_unregister_mapping_type: dereg %s\n", me->name);
	return 0;
}

/* NAT rule tables are ordered, similar to routing tables: most
 * specific to least specific.  I've tried (and maybe failed) to be
 * intuitive here.  Higher the better.  This is a partial ordering.  */
static u_int32_t ip_nat_precedence(const struct ip_nat_pktinfo *mask)
{
	unsigned int i;
	u_int32_t prec = 0;

	/* We build this up most important to least. */

	/* Do we care about device?: 1 bit */
	prec |= mask->device ? 1 : 0;

	/* Source IP bits + destination IP bits: 7 bits */
	prec <<= 7;
	prec |= hweight32(mask->tuplepr.tuple.src.ip)
		+ hweight32(mask->tuplepr.tuple.dst.ip);

	/* Do we care about protocol?: 1 bit */
	prec <<= 1;
	prec |= mask->tuplepr.proto ? 1 : 0;

	/* Per-protocol information: 1+5n bits */
	{
		u_int32_t perprotobits = 1;
		u_int32_t perprotoprec = 0;

		for (i = 0;
		     i < sizeof(mask->tuplepr.tuple.src.u)/sizeof(u_int32_t);
		     i++) {
			perprotoprec
				+= hweight32(mask->tuplepr.tuple.src.u.all[i]);
			perprotobits += 5;
		}

		for (i = 0;
		     i < sizeof(mask->tuplepr.tuple.dst.u)/sizeof(u_int32_t);
		     i++) {
			perprotoprec
				+= hweight32(mask->tuplepr.tuple.dst.u.all[i]);
			perprotobits += 5;
		}

		prec <<= perprotobits;
		prec |= perprotoprec;
	}

	return prec;
}

static int prec_cmpfn(const void *new, const void *curr)
{
	return ((struct ip_nat_rule *)new)->elem.precedence
		> ((struct ip_nat_rule *)curr)->elem.precedence;
}

/* Return 0 or -errno */
static int rule_user_to_rule(const struct ip_nat_rule_user *user,
			     struct ip_nat_rule **rule,
			     size_t size)
{
	unsigned int numtypes, i;
	const char *typename;
	struct ip_nat_multi_range *mr;

	if (user->hook != NF_IP_LOCAL_OUT
	    && user->hook != NF_IP_POST_ROUTING
	    && user->hook != NF_IP_PRE_ROUTING) {
		DEBUGP("rule_user_to_rule: bad hook `%u'\n", user->hook);
		return -EINVAL;
	}

	if (~user->mask.hook) {
		DEBUGP("rule_user_to_rule: bad hook mask `%08X'\n",
		       user->mask.hook);
		return -EINVAL;
	}

	/* Count the number of types. */
	for (typename = (char *)(user+1), numtypes = 0;
	     *typename;
	     typename += strlen(typename)+1, numtypes++)
		DEBUGP("rule_user_to_rule: typename %i `%s'\n",
		       numtypes, typename);

	if (numtypes == 0) {
		DEBUGP("rule_user_to_rule: no types!\n");
		return -EINVAL;
	}

	if (!kmalloc_multi(GFP_USER,
			   rule,
			   sizeof(struct ip_nat_rule)
			   + (numtypes+1)*sizeof(struct ip_nat_mapping_type),
			   &mr,
			   sizeof(struct ip_nat_multi_range)
			   + sizeof(struct ip_nat_range),
			   (void *)NULL))
		return -ENOMEM;

	(*rule)->next = NULL;
	(*rule)->elem.mask = user->mask;
	(*rule)->elem.precedence = ip_nat_precedence(&user->mask);
	memcpy((*rule)->elem.ifname, user->ifname,
	       sizeof((*rule)->elem.ifname));

	(*rule)->elem.info = ((struct ip_nat_pktinfo)
			      { dev_get(user->ifname),
				user->hook,
				user->tupleproto });

	/* Set up range. */
	mr->rangesize = 1;
	mr->range[0] = user->range;
	(*rule)->mr = mr;

	/* Find types, and insert them. */
	for (typename = (char *)(user+1), i = 0;
	     *typename;
	     typename += strlen(typename)+1, i++) {
		DEBUGP("Searching for type `%s'\n", typename);
		if (!((*rule)->elem.type[i] = find_mapping_type(typename, 1))){
			DEBUGP("rule_user_to_rule: can't find type `%s'\n",
			       typename);
			kfree(*rule);
			return -ENOENT;
		}
	}
	(*rule)->elem.type[numtypes] = NULL;

	return 0;
}

/* FIXME: insert should add to range if rule already exists, and
   delete should delete from range. --RR */
static int insert_rule(struct ip_nat_rule_user *ruleuser, size_t len)
{
	int ret;
	struct ip_nat_rule *rule;

	/* double NUL terminate, in case it's bogus. */
	((char *)ruleuser)[len] = ((char *)ruleuser)[len+1] = '\0';
	
	ret = rule_user_to_rule(ruleuser, &rule, len);
	if (ret == 0) {
		unsigned int i;

		modreg_insert(&conns.rule_table, rule, prec_cmpfn);

		for (i = 0; rule->elem.type[i]; i++) {
			if (!rule->elem.type[i]->check(rule, i)) {
				DEBUGP("rule_user_to_rule: type `%s' upset\n",
				       rule->elem.type[i]->name);
				kfree(rule);
				return -EINVAL;
			}
		}

		for (i = 0; rule->elem.type[i]; i++) {
			rule->elem.type[i]->usage++;
			if (rule->elem.type[i]->me)
				__MOD_INC_USE_COUNT(rule->elem.type[i]->me);
		}
	}
	return ret;
}

/* FIXME: Shouldn't compare the holes in the structure. --RR */
static int rule_cmpfn(const void *find, const void *curr)
{
	size_t offset = offsetof(struct ip_nat_rule, elem.info.tuplepr);

	/* Compare everything after device ptr in elem. */
	return memcmp(find + offset, curr + offset,
		      sizeof(struct ip_nat_rule) - offset) == 0;
}

static void delete_rule(const struct ip_nat_rule *rule)
{
	unsigned int i;

	/* Blow up all the bindings created by this rule. */
	delete_bindings(&rule->elem);

	modreg_delete(&conns.rule_table, rule);

	for (i = 0; rule->elem.type[i]; i++) {
		rule->elem.type[i]->usage--;
		if (rule->elem.type[i]->me)
			__MOD_DEC_USE_COUNT(rule->elem.type[i]->me);
	}
	kfree(rule);
}

static int delete_rule_user(const struct ip_nat_rule_user *ruleuser,
			    size_t len)
{
	int ret;
	struct ip_nat_rule *rule;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);

	ret = rule_user_to_rule(ruleuser, &rule, len);
	if (ret == 0) {
		struct ip_nat_rule *match;
		match = modreg_find(&conns.rule_table, rule, rule_cmpfn);
		if (!match)
			ret = -ENOENT;
		else
			delete_rule(match);

		kfree(rule);
	}
	spin_unlock_irqrestore(&ip_nat_lock, flags);
	return ret;
}

static void ipnat_rule_down_device(const struct device *dev)
{
	struct ip_nat_rule *rule;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);
	for (rule = conns.rule_table; rule; rule = rule->next) {
		if (rule->elem.info.device == dev)
			rule->elem.info.device = NULL;
	}
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}
	
static void ipnat_rule_up_device(const struct device *dev, const char *name)
{
	struct ip_nat_rule *rule;
	unsigned long flags;

	DEBUGP("Bringing up device %s (%p).\n", name, dev);
	spin_lock_irqsave(&ip_nat_lock, flags);
	for (rule = conns.rule_table; rule; rule = rule->next) {
		if (strcmp(rule->elem.ifname, name) == 0)
			rule->elem.info.device = dev;
	}
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}

static int do_ip_nat_ctl(int cmd, void *user, unsigned int len)
{
	int ret;

	switch (cmd) {
	case IP_NAT_SO_DELETE:
	case IP_NAT_SO_INSERT: {
		struct ip_nat_rule_user *ruleuser;

		if (len < sizeof(struct ip_nat_rule)) {
			DEBUGP("IP_NAT_SO_INSERT/DELETE: len %u < %u\n",
			       len, sizeof(struct ip_nat_rule));
			ret = -EINVAL;
		} else if (!(ruleuser = kmalloc(len+2, GFP_USER)))
			ret = -ENOMEM;
		else if (copy_from_user(ruleuser, user, len)) {
			kfree(ruleuser);
			ret = -EFAULT;
		} else {
			ret = (cmd == IP_NAT_SO_INSERT ?
			       insert_rule(ruleuser, len)
			       : delete_rule_user(ruleuser, len));
			kfree(ruleuser);
		}
		break;
	}

	default:
		panic("ip_nat_rule: Someone wake Rusty.\n");
	}

	return ret;
}

#ifdef CONFIG_PROC_FS
static int dump_rule(char *buffer, const struct ip_nat_rule *rule)
{
	int len = 0;
	unsigned int i;

	len = sprintf(buffer + len,
		      "%s %16s %u.%u.%u.%u/%lu -> %u.%u.%u.%u/%lu ",
		      hooknames[rule->elem.info.hook],
		      rule->elem.ifname,
		      IP_PARTS(rule->elem.info.tuplepr.tuple.src.ip),
		      rule->elem.mask.tuplepr.tuple.src.ip == 0xFFFFFFFF ? 32
		      : ffz(rule->elem.mask.tuplepr.tuple.src.ip),
		      IP_PARTS(rule->elem.info.tuplepr.tuple.dst.ip),
		      rule->elem.mask.tuplepr.tuple.dst.ip == 0xFFFFFFFF ? 32
		      : ffz(rule->elem.mask.tuplepr.tuple.dst.ip));

	if (rule->elem.mask.tuplepr.proto)
		len += sprintf(buffer + len, "proto=%u ",
			       rule->elem.info.tuplepr.proto);

	for (i = 0; rule->elem.type[i]; i++)
		len += sprintf(buffer + len, ":%s", rule->elem.type[i]->name);
	len += sprintf(buffer + len, "\n");

	return len;
}

/* List of rules; offset in records not bytes. */
static int nat_get_rules(char *buffer, char **start, off_t offset,
			 int length, int dummy)
{
	struct ip_nat_rule *rule;
	off_t upto = 0;
	unsigned long flags;
	int len = 0,last_len = 0;

	spin_lock_irqsave(&ip_nat_lock, flags);
	for (rule = conns.rule_table; rule; rule = rule->next, upto++) {
		if (upto < offset) continue;

		last_len = len;
		len += dump_rule(buffer+len, rule);
		if (len > length) {
			len = last_len;
			break;
		}
	}

	spin_unlock_irqrestore(&ip_nat_lock, flags);
	buffer[len] = '\0';

	DEBUGP("nat_get_rules: Length = %i (of %i).  Offset = %li.\n",
	       len, length, upto);
	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}
#endif

static struct nf_setsockopt_ops ip_nat_sockopts
= { NULL, PF_INET, IP_NAT_BASE_CTL, IP_NAT_SO_MAX, do_ip_nat_ctl };

static int init_ip_nat_rule(void)
{
	int ret;

	ret = ip_nat_register_mapping_type(&mapping_generic);
	if (ret == 0) {
		/* Generic doesn't count as usage. */
		MOD_DEC_USE_COUNT;
		ret = nf_register_sockopt(&ip_nat_sockopts);
		if (ret != 0)
			ip_nat_unregister_mapping_type(&mapping_generic);
	}
	return ret;
}

static void cleanup_ip_nat_rule(void)
{
	nf_unregister_sockopt(&ip_nat_sockopts);

	/* Delete all the rules. */
	while (conns.rule_table)
		modreg_delete(&conns.rule_table, conns.rule_table);

	/* Deregistering generic will dec. usage count. */
	MOD_INC_USE_COUNT;
	while (mapping_types)
		ip_nat_unregister_mapping_type(mapping_types);
}

