/*
 * This is the 1999 rewrite of IP Firewalling, aiming for kernel 2.3.x.
 *
 * Copyright (C) 1999 Paul F. Russell & Michael J. Neuling
 */
#ifdef __KERNEL__
#include <linux/config.h>

#include <asm/smp.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/minmax.h>
#include <linux/modreglist.h>
#include <linux/kmod.h>
#include <linux/stat.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/netdevice.h>
#else
#include "ip_tables_usersupp.h"
#endif

#include "ip_tables.h"
#include "ip_tables_standard.h"

#ifndef IP_OFFSET
#define IP_OFFSET 0x1FFF
#endif

/* Understanding locking in this code: (thanks to Alan Cox for using
 * little words to explain this to me). -- PR
 *
 * In UP, there can be two packets traversing the chains:
 * 1) A packet from the current userspace context
 * 2) A packet off the bh handlers (timer or net).
 *
 * For SMP (kernel v2.1+), multiply this by # CPUs.
 *
 * [Note that this in not correct for 2.2 - because the socket code always
 *  uses lock_kernel() to serialize, and bottom halves (timers and net_bhs)
 *  only run on one CPU at a time.  This will probably change for 2.3.
 *  It is still good to use spinlocks because that avoids the global cli()
 *  for updating the tables, which is rather costly in SMP kernels -AK]
 *
 * This means counters and backchains can get corrupted if no precautions
 * are taken.
 *
 * To actually alter a chain on UP, we need only do a cli(), as this will
 * stop a bh handler firing, as we are in the current userspace context
 * (coming from a setsockopt()).
 *
 * On SMP, we need a write_lock_irqsave(), which is a simple cli() in
 * UP.
 *
 * For backchains and counters, we use an array, indexed by
 * [cpu_number_map[smp_processor_id()]*2 + !in_interrupt()]; the array
 * is of size [smp_num_cpus*2].  For v2.0, smp_num_cpus is effectively
 * 1.  So, confident of uniqueness, we modify counters even though we
 * only have a read lock (to read the counters, you need a write lock,
 * though).  */

/* Why I didn't use straight locking... -- PR
 *
 * The backchains can be separated out of the ip_chains structure, and
 * allocated as needed inside ipt_check().
 *
 * The counters, however, can't.  Trying to lock these means blocking
 * interrupts every time we want to access them.  This would suck HARD
 * performance-wise.  Not locking them leads to possible corruption,
 * made worse on 32-bit machines (counters are 64-bit).  */

/* A few tricks.  We avoid running off the end by keeping a special entry
 * there which uses ipt_error as its target function.  We avoid underflow
 * (more IPT_RETURN than jumps) by pointing the back pointer from the
 * entry points at this entry too.
 */

#define DEBUG_IP_FIREWALL
/*#define DEBUG_ALLOW_ALL*/ /* Useful for remote debugging */
#define DEBUG_IP_FIREWALL_USER
#define DEBUG_IP_FIREWALL_LOCKING
#define DEBUG

#ifdef CONFIG_SMP
#define SLOT_NUMBER() (cpu_number_map[smp_processor_id()]*2 + !in_interrupt())
#else
#define SLOT_NUMBER() (!in_interrupt())
#endif
#define NUM_SLOTS (smp_num_cpus*2)

#if defined(DEBUG_IP_FIREWALL) || defined (DEBUG_IP_FIREWALL_USER)
/* All the better to debug you with... */
#define static
#endif
#define static

#ifdef DEBUG_IP_FIREWALL_LOCKING
static unsigned int iptd_rlocks, iptd_wlocks;
#define IPTD_DEBUG_LOCK(d)			\
do {						\
	IPTD_DONT_HAVE_LOCK(d);			\
	d |= (1 << SLOT_NUMBER());		\
} while (0)

#define IPTD_DEBUG_UNLOCK(d)			\
do {						\
	IPTD_HAVE_LOCK(d);			\
	d &= ~(1 << SLOT_NUMBER());		\
} while (0)

#define IPTD_DONT_HAVE_LOCK(d)					\
do {								\
	if ((d) & (1 << SLOT_NUMBER()))				\
		printk("%s:%i: Got lock on %i already!\n", 	\
		       __FILE__, __LINE__, SLOT_NUMBER());	\
} while(0)

#define IPTD_HAVE_LOCK(d)				\
do {							\
	if (!((d) & (1 << SLOT_NUMBER())))		\
	printk("%s:%i:No lock on %i!\n", 		\
	       __FILE__, __LINE__, SLOT_NUMBER());	\
} while (0)

#else
#define IPTD_DEBUG_LOCK(d) do { } while(0)
#define IPTD_DEBUG_UNLOCK(d) do { } while(0)
#define IPTD_DONT_HAVE_LOCK(d) do { } while(0)
#define IPTD_HAVE_LOCK(d) do { } while(0)
#endif /*DEBUG_IP_FIREWALL_LOCKING*/

#define IPTD_READ_LOCK(l) do { IPTD_DEBUG_LOCK(iptd_rlocks); read_lock(l); } while (0)
#define IPTD_READ_UNLOCK(l) do { IPTD_DEBUG_UNLOCK(iptd_rlocks); read_unlock(l); } while (0)
#define IPTD_WRITE_LOCK_IRQ(l,f) do { IPTD_DEBUG_LOCK(iptd_wlocks); write_lock_irqsave(l,f); } while (0)
#define IPTD_WRITE_UNLOCK_IRQ(l,f) do { IPTD_DEBUG_UNLOCK(iptd_wlocks); write_unlock_irqrestore(l,f); } while (0)

/*
 *	Implement IP packet firewall
 */

#ifdef DEBUG_IP_FIREWALL
#define dprintf(format, args...)  printk(format , ## args)
#else
#define dprintf(format, args...)
#endif

#ifdef DEBUG_IP_FIREWALL_USER
#define duprintf(format, args...) printk(format , ## args)
#else
#define duprintf(format, args...)
#endif

/* Lock around ipt_table_info table */
rwlock_t ipt_lock = RW_LOCK_UNLOCKED;

struct ipt_kern_entry
{
	struct ipt_match *match;
	struct ipt_ip ip;
	union ipt_matchinfo matchinfo;
	struct ipt_target *act;
	unsigned int cacheinfo;
	union ipt_targinfo targinfo;
};

struct per_slot
{
	unsigned int backpointer;
	struct ipt_counters counters;
};

/* The table itself */
struct ipt_table_info
{
	unsigned int number;
	unsigned int hook_entry[3];
	/* number of these */
	struct ipt_kern_entry *table;

	/* Actually NUM_SLOTS of these, each one pointing to [number] */
	struct per_slot *slots[0];
};

static struct ipt_table_info *ipt_table_info;
struct ipt_target *ipt_target;
struct ipt_match *ipt_match;

#define ADD_COUNTER(c,b,p) do { (c).bcnt += (b); (c).pcnt += (p); } while(0)

/* Returns whether matches rule or not. */
static int
ip_packet_match(const struct iphdr *ip,
		const char *indev,
		const char *outdev,
		const struct ipt_ip *ipinfo,
		int isfrag)
{
#define FWINV(bool,invflg) ((bool) ^ !!(ipinfo->invflags & invflg))
	/*
	 *	This is a bit simpler as we don't have to walk
	 *	an interface chain as you do in BSD - same logic
	 *	however.
	 */

	if (FWINV((ip->saddr&ipinfo->smsk.s_addr) != ipinfo->src.s_addr,
		  IPT_INV_SRCIP)
	    || FWINV((ip->daddr&ipinfo->dmsk.s_addr) != ipinfo->dst.s_addr,
		     IPT_INV_DSTIP)) {
		dprintf("Source or dest mismatch.\n");

		dprintf("SRC: %u. Mask: %u. Target: %u.%s\n", ip->saddr,
			ipinfo->smsk.s_addr, ipinfo->src.s_addr,
			ipinfo->invflags & IPT_INV_SRCIP ? " (INV)" : "");
		dprintf("DST: %u. Mask: %u. Target: %u.%s\n", ip->daddr,
			ipinfo->dmsk.s_addr, ipinfo->dst.s_addr,
			ipinfo->invflags & IPT_INV_DSTIP ? " (INV)" : "");
		return 0;
	}

	/* Look for ifname match */
	if (FWINV(memcmp(indev, ipinfo->iniface, ipinfo->iniface_len) != 0,
		  IPT_INV_VIA_IN)) {
		dprintf("VIA in mismatch (%s vs %s).%s\n",
			indev, ipinfo->iniface,
			ipinfo->invflags&IPT_INV_VIA_IN ?" (INV)":"");
		return 0;
	}
	if (FWINV(memcmp(outdev, ipinfo->outiface, ipinfo->outiface_len) != 0,
		  IPT_INV_VIA_OUT)) {
		dprintf("VIA out mismatch (%s vs %s).%s\n",
			outdev, ipinfo->outiface,
			ipinfo->invflags&IPT_INV_VIA_OUT?" (INV)":"");
		return 0;
	}

	/* If we have a fragment rule but the packet is not a fragment
	 * the we return zero */
	if (FWINV((ipinfo->flags&IPT_F_FRAG) && !isfrag, IPT_INV_FRAG)) {
		dprintf("Fragment rule but not fragment.%s\n",
			ipinfo->invflags & IPT_INV_FRAG ? " (INV)" : "");
		return 0;
	}

	/* Check TOS match */
	if ((ipinfo->flags & IPT_F_TOS)
	    && FWINV(ipinfo->tos != ip->tos, IPT_INV_TOS)) {
		dprintf("Packet TOS %u does not match %u.%s\n",
			ip->tos, ipinfo->tos,
			ipinfo->invflags&IPT_INV_TOS ? " (INV)":"");
		return 0;
	}

	/* Check specific protocol */
	if (ipinfo->proto
	    && FWINV(ip->protocol != ipinfo->proto, IPT_INV_PROTO)) {
		dprintf("Packet protocol %hi does not match %hi.%s\n",
			ip->protocol, ipinfo->proto,
			ipinfo->invflags&IPT_INV_PROTO ? " (INV)":"");
		return 0;
	}

/* This should be a separate add-on --RR */
#if 0
	/* Look for IP options match */
	if (ipinfo->option) {
		const u_int8_t *opt = (u_int8_t *)(ip+1);
		const u_int8_t *end = opt + ip->ihl*4;
		int found = 0;

		/* FIXME: Run 0-length and > ip->ihl IP options
		 * against various stacks.  Might be fun. --RR */
		while (opt < end && !found) {
			found = (ipinfo->option == *opt);
			if (*opt < 2) opt++;
			else opt += opt[1]?:1;
		}

		if (FWINV(!found, IPT_INV_OPTION)) {
			dprintf("Option %u%s found\n", ipinfo->option,
			ipinfo->invflags&IPT_INV_OPTION ? "" : " not");
			return 0;
		}
	}			
#endif

	return 1;
}

static int ip_checkentry(const struct ipt_ip *ip)
{
	if (ip->flags & ~IPT_F_MASK) {
		duprintf("Unknown flag bits set: %08X\n",
			 ip->flags & ~IPT_F_MASK);
		return 0;
	}
	if (ip->invflags & ~IPT_INV_MASK) {
		duprintf("Unknown invflag bits set: %08X\n",
			 ip->invflags & ~IPT_INV_MASK);
		return 0;
	}
	return 1;
}

static int
ipt_standard_target(struct sk_buff *skb,
		    const char *indev,
		    const char *outdev,
		    const union ipt_targinfo *info,
		    unsigned int pos)
{
	if (info->standard.mark) {
		skb->nfreason = NF_SPECIAL_SET_BY_IPTABLES;
		skb->nfmark = info->standard.mark;
	}

	return info->standard.verdict;
}

static int
ipt_error(struct sk_buff *skb,
	  const char *indev,
	  const char *outdev,
	  const union ipt_targinfo *info,
	  unsigned int pos)
{
	if (net_ratelimit())
		printk("ip_tables: error: `%s'\n", info->error);

	return -NF_DROP - 1;
}

/*
 *	Returns one of the generic firewall policies, like FW_ACCEPT.
 */
static unsigned int
ipt_check(struct sk_buff *skb,
	  unsigned int pos,
	  const char *indev,
	  const char *outdev,
	  unsigned int *cache)
{
	u_int16_t offset;
	struct iphdr *ip = skb->nh.iph;
	void *protohdr = (u_int32_t *)ip + ip->ihl;
	u_int16_t datalen = skb->len - ip->ihl * 4;
	unsigned int i = 0;
	int spos = (int)pos;
	const unsigned int number = ipt_table_info->number;

	/* We handle fragments by dealing with the first fragment as
	 * if it was a normal packet.  All other fragments are treated
	 * normally, except that they will NEVER match rules that ask
	 * things we don't know, ie. tcp syn flag or ports).  If the
	 * rule is also a fragment-specific rule, non-fragments won't
	 * match it. */
	offset = ntohs(ip->frag_off) & IP_OFFSET;

	/* Special magic for TCP: to quote Alan:
	     Don't allow a fragment of TCP 8 bytes in. Nobody normal
	     causes this. Its a cracker trying to break in by doing a
	     flag overwrite to pass the direction checks.  */
	if (offset == 1 && ip->protocol == IPPROTO_TCP) {
		duprintf("Dropping evil TCP frag.\n");
		return NF_DROP;
	}

	while (1) {
		unsigned int oldpos = (unsigned int)spos;
		struct ipt_kern_entry *e = &ipt_table_info->table[oldpos];
		*cache |= e->cacheinfo;
		/* We call the specific match first; it's most likely
		   to eliminate. */
		if ((!e->match->match
		     || e->match->match(skb, indev, outdev, &e->matchinfo,
					offset, protohdr, datalen))
		    && ip_packet_match(skb->nh.iph, indev, outdev, &e->ip,
				       offset)) {
			struct per_slot *slots
				= ipt_table_info->slots[SLOT_NUMBER()];
			dprintf("We have a match on %u.\n", oldpos);
			ADD_COUNTER(slots[oldpos].counters,
				    ntohs(ip->tot_len), 1);

			spos = e->act->target(skb, indev, outdev, &e->targinfo,
					      spos);

			/* Pop from stack? */
			if (spos == -IPT_RETURN - 1) {
				dprintf("ipt_check: pop %u.\n", oldpos);
				spos = slots[oldpos].backpointer;
			}
			/* A verdict. */
			else if (spos < 0) break;
			/* Push onto stack? */
			else if (spos != (int)oldpos+1) {
				slots[spos].backpointer = oldpos;
				/* Overflow? */
				if (++i > number) {
					dprintf("ip_tables: Overflow!\n");
					spos = ipt_error(skb, indev, outdev,
							 NULL, spos);
				}
			}
		}
		else spos++;
	}

#ifdef DEBUG_ALLOW_ALL
	return NF_ACCEPT;
#else
	return (unsigned int)(-spos - 1);
#endif
}

static void *
find_inlist(void *headaddr, const char *name)
{
	void *ret;

	duprintf("find_inlist: searching for `%s' in %s.\n",
		 name, headaddr == &ipt_target ? "ipt_target"
		 : headaddr == &ipt_match ? "ipt_match" : "UNKNOWN");

	ret = modreg_named_find(headaddr, name);
	if (!ret) {
		char modulename[IPT_FUNCTION_MAXNAMELEN + sizeof("ipt_")];
		strcpy(modulename, "ipt_");
		strcat(modulename, name);
		duprintf("find_inlist: loading `%s'.\n", modulename);
		request_module(modulename);
		ret = modreg_named_find(headaddr, name);
	}
	return ret;
}

/* Checks the user-supplied table segment. */
static int
translate_table(struct ipt_table_info *newinfo,
			   const struct ipt_replace *repl)
{
	unsigned int i;
	struct per_slot *newslots = (void *)newinfo
		+ sizeof(struct ipt_table_info)
		+ NUM_SLOTS * sizeof(struct per_slot *)
		+ repl->number * sizeof(struct ipt_kern_entry);

	duprintf("translate_table: newslots = %p.\n", newslots);

	/* Check hooks. */
	for (i = 0; i < sizeof(repl->hook_entry)/sizeof(unsigned int); i++) {
		if (repl->hook_entry[i] > repl->number) {
			duprintf("Bad hook %u->%u.\n", i, repl->hook_entry[i]);
			return 0;
		}
	}

	duprintf("translate_table: hooks OK.\n");
	*newinfo = ((struct ipt_table_info)
		    { repl->number,
		      { repl->hook_entry[0],
			repl->hook_entry[1],
			repl->hook_entry[2] },
		      (void *)newinfo
		      + sizeof(struct ipt_table_info)
		      + NUM_SLOTS * sizeof(struct per_slot *),
		      { } });

	duprintf("translate_table: newinfo assigned.\n");
	/* Set up pointers to slots, slot backchains to point to last
           node, and zero counters */
	for (i = 0; i < NUM_SLOTS; i++) {
		unsigned int j;

		newinfo->slots[i] = newslots + i*newinfo->number;
		for (j = 0; j < newinfo->number; j++) {
			duprintf("translate_table:"
				 " setting up perslot[%i][%i] (%p).\n",
				 i, j, &newinfo->slots[i][j]);
			newinfo->slots[i][j]
				= ((struct per_slot) { newinfo->number-1,
						       { 0, 0 } });
		duprintf("packets = %u|%u bytes = %u|%u\n",
			 (unsigned int)(newinfo->slots[i][j].counters.pcnt >> 32),
			 (unsigned int)newinfo->slots[i][j].counters.pcnt,
			 (unsigned int)(newinfo->slots[i][j].counters.bcnt >> 32),
			 (unsigned int)newinfo->slots[i][j].counters.bcnt);
		}
	}

	/* Each target name must be valid.  Each sanity check must pass. */
	for (i = 0; i < repl->number; i++) {
		const struct ipt_entry *e = &repl->entry_array[i];
		struct ipt_target *act;
		struct ipt_match *match;

		duprintf("translate_table: dealing with entry %i %p->%p.\n",
			 i, &newinfo->table[i], &newinfo->table[i+1]);

		act = find_inlist(&ipt_target, e->target_name);
		if (!act) {
			duprintf("ip_tables: no target by name `%s'.\n",
				 e->target_name);
			return 0;
		}

		match = find_inlist(&ipt_match, e->match_name);
		if (!match) {
			duprintf("ip_tables: no match by name `%s'.\n",
				 e->match_name);
			return 0;
		}

		if (act->checkentry
		    && !act->checkentry(&e->targinfo, i, newinfo->number)) {
			duprintf("ip_tables: check failed for `%s'.\n",
				 e->target_name);
			return 0;
		}
		if ((match->checkentry
		     && !match->checkentry(&e->ip, &e->matchinfo))
		    || !ip_checkentry(&e->ip)) {
			duprintf("ip_tables: check failed for `%s'.\n",
				 e->match_name);
			return 0;
		}
		newinfo->table[i]
			= ((struct ipt_kern_entry)
			   { match, e->ip, e->matchinfo, act, e->nfcache,
			     e->targinfo });

		/* FIXME: Do we need this? --RR */
		/* Set counters in the first slot. */
		newinfo->slots[0][i].counters = e->counters;
		duprintf("(%p) packets = %u|%u bytes = %u|%u\n",
			 &newinfo->slots[0][i],
			 (unsigned int)(newinfo->slots[0][i].counters.pcnt >> 32),
			 (unsigned int)newinfo->slots[0][i].counters.pcnt,
			 (unsigned int)(newinfo->slots[0][i].counters.bcnt >> 32),
			 (unsigned int)newinfo->slots[0][i].counters.bcnt);
	}

	duprintf("Slot[0][0]: packets = %u|%u bytes = %u|%u\n",
		 (unsigned int)(newinfo->slots[0][0].counters.pcnt>>32),
		 (unsigned int)newinfo->slots[0][0].counters.pcnt,
		 (unsigned int)(newinfo->slots[0][0].counters.bcnt>>32),
		 (unsigned int)newinfo->slots[0][0].counters.bcnt);
	return 1;
}

static void
sum_counters(struct ipt_counters *counters,
			 struct per_slot *const *slots,
			 unsigned int offset)
{
	unsigned int i;

	counters->pcnt = counters->bcnt = 0;
	for (i = 0; i < NUM_SLOTS; i++) {
		duprintf("Summing counters in slot[%i][%i] (%p)\n",
			 i, offset, &slots[i][offset]);
		duprintf("packets = %u|%u bytes = %u|%u\n",
			 (unsigned int)(slots[i][offset].counters.pcnt >> 32),
			 (unsigned int)slots[i][offset].counters.pcnt,
			 (unsigned int)(slots[i][offset].counters.bcnt >> 32),
			 (unsigned int)slots[i][offset].counters.bcnt);

		ADD_COUNTER(*counters,
			    slots[i][offset].counters.bcnt,
			    slots[i][offset].counters.pcnt);
	}
}

/* Gets counters. */
static void
get_counters(const struct ipt_table_info *info,
			 struct ipt_counters counters[])
{
	unsigned int i;

#ifdef DEBUG_IP_FIREWALL_LOCKING
	/* Need lock if operating on live data */
	if (info == ipt_table_info)
		IPTD_HAVE_LOCK(iptd_wlocks);
#endif

	for (i = 0; i < info->number; i++)
		sum_counters(&counters[i], info->slots, i);
}

static int
get_entries(struct ipt_get_entries *entries)
{
	unsigned long flags;
	unsigned int i;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	if (entries->num_entries != ipt_table_info->number) {
		IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
		duprintf("get_entries: I've got %u not %u!\n",
			 ipt_table_info->number, entries->num_entries);
		return -EINVAL;
	}
	for (i = 0; i < ipt_table_info->number; i++) {
		/* Translate back to userland friendly. */
		struct ipt_kern_entry *kent = &ipt_table_info->table[i];
		struct ipt_entry *ent = &entries->entries[i];

		ent->ip = kent->ip;
		ent->matchinfo = kent->matchinfo;
		ent->nfcache = kent->cacheinfo;
		memcpy(ent->target_name, kent->act->name,
		       sizeof(ent->target_name));
		memcpy(ent->match_name, kent->match->name,
		       sizeof(ent->match_name));
		duprintf("get_entries: copying target `%s' match `%s'\n",
			 ent->target_name, ent->match_name);
		sum_counters(&ent->counters, ipt_table_info->slots, i);
		ent->targinfo = kent->targinfo;
	}
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
	return 0;
}

static int
add_counters(struct ipt_counters_info *paddc)
{
	unsigned long flags;
	unsigned int i;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	if (ipt_table_info->number != paddc->num_counters) {
		IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
		return -EINVAL;
	}
	for (i = 0; i < ipt_table_info->number; i++) {
		/* We're lazy, and add to the first slot; overflow
		 * works its fey magic and everything is OK. */
		ADD_COUNTER(ipt_table_info->slots[0][i].counters,
			    paddc->counters[i].bcnt,
			    paddc->counters[i].pcnt);
	}
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
	return 0;
}

static struct ipt_table_info *
replace_table(struct ipt_replace *repl, int *error)
{
	unsigned int i;
	struct ipt_table_info *newinfo, *oldinfo;
	unsigned long flags;

	/* Allocate slots and new entries. */
	newinfo = kmalloc(sizeof(struct ipt_table_info)
			  + NUM_SLOTS * sizeof(struct per_slot *)
			  + repl->number * sizeof(struct ipt_kern_entry)
			  + repl->number * NUM_SLOTS * sizeof(struct per_slot),
			  GFP_USER);
	if (!newinfo) {
		*error = -ENOMEM;
		return NULL;
	}

	duprintf("ip_tables: table %p to %p\n", newinfo,
		 (void *)newinfo
		 + sizeof(struct ipt_table_info)
		 + NUM_SLOTS * sizeof(struct per_slot *)
		 + repl->number * sizeof(struct ipt_kern_entry)
		 + repl->number * NUM_SLOTS * sizeof(struct per_slot));

	if (!translate_table(newinfo, repl))
		goto bad_data;

	duprintf("ip_tables: Translated table\n");
	/* Increase module usage counts. */
	for (i = 0; i < newinfo->number; i++) {
		if (newinfo->table[i].act->me)
			__MOD_INC_USE_COUNT(newinfo->table[i].act->me);
	}

	duprintf("Slot[0][0]: packets = %u|%u bytes = %u|%u\n",
		 (unsigned int)(newinfo->slots[0][0].counters.pcnt>>32),
		 (unsigned int)newinfo->slots[0][0].counters.pcnt,
		 (unsigned int)(newinfo->slots[0][0].counters.bcnt>>32),
		 (unsigned int)newinfo->slots[0][0].counters.bcnt);

	/* Do the substitution. */
	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	/* Check inside lock: is the number of counters correct? */
	if (repl->num_counters != ipt_table_info->number) {
		IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
		goto bad_data;
	}
	oldinfo = ipt_table_info;
	ipt_table_info = newinfo;
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);

	return oldinfo;

 bad_data:
	kfree(newinfo);
	*error = -EINVAL;
	return NULL;
}

static void *
alloc_from_user(size_t supposed_len_base,
			     size_t size_per_entry,
			     void *user, unsigned int len,
			     int *error)
{
	unsigned int number;
	void *ret;

	if ((*error = get_user(number, (unsigned int *)user)) != 0)
		return NULL;

	if (supposed_len_base + number * size_per_entry != len) {
		duprintf("ip_tables: bad length (%u = %u + %u * %u not %u).\n",
			 supposed_len_base + number * size_per_entry,
			 supposed_len_base, number, size_per_entry, len);
		*error = -EINVAL;
		return NULL;
	}
	ret = kmalloc(len, GFP_USER);
	if (!ret) {
		duprintf("ip_tables: oom (%u).\n", len);
		*error = -ENOMEM;
	}
	*((unsigned int *)ret) = number;
	return ret;
}

static void *
var_from_user(size_t supposed_len_base,
			   size_t size_per_entry,
			   void *user, unsigned int len,
			   int *error)
{
	void *ret;

	if ((ret = alloc_from_user(supposed_len_base, size_per_entry, user,
				   len, error)) != NULL) {
		*error = copy_from_user(ret, user, len);
		if (*error != 0) {
			kfree(ret);
			ret = NULL;
		}
	}
	return ret;
}

static int
do_replace(void *user, unsigned int len)
{
	int ret;
	struct ipt_replace *repl;

	repl = var_from_user(sizeof(struct ipt_replace),
			     sizeof(struct ipt_entry),
			     user, len, &ret);
	if (repl) {
		struct ipt_counters *uout = repl->counters;
		repl->counters
			= kmalloc(repl->num_counters
				  * sizeof(struct ipt_counters),
				  GFP_USER);
		if (!repl->counters)
			ret = -ENOMEM;
		else {
			struct ipt_table_info *oldinfo;

			oldinfo = replace_table(repl, &ret);
			if (oldinfo) {
				unsigned int i;
				/* Get the old counters. */
				get_counters(oldinfo, repl->counters);

				/* Decrease module usage counts */
				for (i = 0; i < oldinfo->number; i++) {
					if (oldinfo->table[i].act->me)
						__MOD_DEC_USE_COUNT(oldinfo
								    ->table[i]
								    .act->me);
				}
				kfree(oldinfo);
				copy_to_user(uout, repl->counters,
					     sizeof(struct ipt_counters)
					     * repl->num_counters);
			}
			kfree(repl->counters);
		}
		kfree(repl);
	}
	return 0;
}

static int
do_ipt_ctl(int cmd, void *user, unsigned int len)
{
	int ret;

	switch (cmd) {
	case IPT_SO_GET_INFO:
		if (len != sizeof(struct ipt_getinfo))
			ret = -EINVAL;
		else {
			struct ipt_getinfo info
				= { { 0, 0, 0 }, ipt_table_info->number };
			memcpy(info.hook_entry, ipt_table_info->hook_entry,
			       sizeof(info.hook_entry));
			ret = copy_to_user(user, &info, len);
		}
		break;

	case IPT_SO_REPLACE:
		ret = do_replace(user, len);
		break;

	case IPT_SO_ADD_COUNTERS: {
		struct ipt_counters_info *pgetc
			= var_from_user(sizeof(struct ipt_counters_info),
					sizeof(struct ipt_counters),
					user, len, &ret);
		if (pgetc) {
			ret = add_counters(pgetc);
			kfree(pgetc);
		}
		break;
	}

	case IPT_SO_GET_ENTRIES: {
		struct ipt_get_entries *entries
			= alloc_from_user(sizeof(struct ipt_get_entries),
					  sizeof(struct ipt_entry),
					  user, len, &ret);
		if (entries) {
			if ((ret = get_entries(entries)) == 0)
				ret = copy_to_user(user, entries,
						   ipt_table_info->number
						   * sizeof(struct ipt_entry));
			kfree(entries);
		}
		break;
	}

	default:
		duprintf("do_ipt_ctl:  unknown request %i\n", cmd);
		ret = -EINVAL;
	}

	return ret;
}

/* The work comes in here from netfilter.c. */
static unsigned int
ipt_hook(unsigned int hook,
			     struct sk_buff **pskb,
			     const struct device *indev,
			     const struct device *outdev)
{
	unsigned int ret;

	dprintf("ipt_hook: hook %u.\n", hook);
	/* Let locally-generated evil packets through. */
	if (hook == NF_IP_LOCAL_OUT
	    && ((*pskb)->nh.iph->ihl * 4 < sizeof(struct iphdr)
		|| (*pskb)->len < sizeof(struct iphdr))) {
		dprintf("ipt_hook: happy cracking.\n");
		return NF_ACCEPT;
	}

	IPTD_READ_LOCK(&ipt_lock);
	ret = ipt_check(*pskb, ipt_table_info->hook_entry[hook-NF_IP_LOCAL_IN],
			indev ? indev->name : "",
			outdev ? outdev->name : "",
			&(*pskb)->nfcache);
	IPTD_READ_UNLOCK(&ipt_lock);
	dprintf("ipt_hook: returning %u.\n", ret);

	return ret;
}

/* Registration hooks for targets. */
int
ipt_register_target(struct ipt_target *target)
{
	if (modreg_named_insert(&ipt_target, target)) {
		MOD_INC_USE_COUNT;
		return 0;
	}

		duprintf("ipt_register_target: `%s' already in list!\n",
			 target->name);
		return -EINVAL;
	}

int
ipt_unregister_target(struct ipt_target *target)
{
#ifdef DEBUG_IP_FIREWALL_USER
	if (!modreg_inlist(&ipt_target, target)) {
		printk("ipt_unregister_target: `%s' not in list!\n",
		       target->name);
		return -EINVAL;
	}
#endif
	modreg_delete(&ipt_target, target);
	MOD_DEC_USE_COUNT;
	return 0;
}

int
ipt_register_match(struct ipt_match *match)
{
	if (modreg_named_find(&ipt_match, match->name))
		duprintf("ipt_register_match: `%s' already in list.\n",
			 match->name);

	if (modreg_named_insert(&ipt_match, match)) {
		MOD_INC_USE_COUNT;
		return 0;
	}

		duprintf("ipt_register_match: `%s' already in list!\n",
			 match->name);
		return -EINVAL;
	}

int
ipt_unregister_match(struct ipt_match *match)
{
#ifdef DEBUG_IP_FIREWALL_USER
	if (!modreg_inlist(&ipt_match, match)) {
		printk("ipt_unregister_match: `%s' not in list!\n",
		       match->name);
		return -EINVAL;
	}
#endif
	modreg_delete(&ipt_match, match);
	MOD_DEC_USE_COUNT;
	return 0;
}

static int
ipt_standard_checkentry(const union ipt_targinfo *info,
				   unsigned int pos,
				   unsigned int finalpos)
{
	/* Check standard info. */
	if (info->standard.verdict < -IPT_RETURN - 1
	    || info->standard.verdict > (int)finalpos) {
		duprintf("ipt_standard_check: bad verdict (%i)\n",
			 info->standard.verdict);
		return 0;
	}
	return 1;
}

static struct nf_hook_ops ipt_ops[]
= { { NULL, ipt_hook, NULL, PF_INET, NF_IP_LOCAL_IN },
    { NULL, ipt_hook, NULL, PF_INET, NF_IP_FORWARD },
    { NULL, ipt_hook, NULL, PF_INET, NF_IP_LOCAL_OUT }
};

/* The built-in targets: error and standard. */
static struct ipt_target ipt_error_reg
= { NULL, IPT_ERROR_TARGET, ipt_error, NULL, NULL };
static struct ipt_target ipt_standard_reg
= { &ipt_error_reg, IPT_STANDARD_TARGET, ipt_standard_target,
    ipt_standard_checkentry, NULL };

/* The built-in match: normal IP match. */
static struct ipt_match ipt_standard_match
= { NULL, IPT_STANDARD_MATCH, NULL, NULL, NULL };
static struct nf_setsockopt_ops ipt_sockopts
= { NULL, PF_INET, IPT_BASE_CTL, IPT_SO_MAX, do_ipt_ctl };

/* ACCEPT rule, forward rule (if module, verdict may be modified on
   module load), ACCEPT rule, and an error rule. */

/* Looking for memcpy?  If this struct changes, could get called here. */
static struct
{
	struct ipt_replace repl;
	struct ipt_entry entries[4];
} initial_table __initdata
= { { 4, { 0, 1, 2 }, 0, NULL, { } },
    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", 0, 0, 0, 0 },
	{ { { 0, 0 }, { 0, 0 }, 0, 0, 0 } },
	0, "", "", { 0, 0 },
	{ { -NF_ACCEPT - 1, 0, { 0, 0 } } } },
      { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", 0, 0, 0, 0 },
	{ { { 0, 0 }, { 0, 0 }, 0, 0, 0 } },
	0, "", "", { 0, 0 },
	{ { -NF_ACCEPT - 1, 0, { 0, 0 } } } },
      { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", 0, 0, 0, 0 },
	{ { { 0, 0 }, { 0, 0 }, 0, 0, 0 } },
	0, "", "", { 0, 0 },
	{ { -NF_ACCEPT - 1, 0, { 0, 0 } } } },
      { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", 0, 0, 0, 0 },
	{ { { 0, 0 }, { 0, 0 }, 0, 0, 0 } },
	0, "", IPT_ERROR_TARGET, { 0, 0 },
	{ { 0, 0, { 0, 0 } } } }
    }
};

static int __init
init(void)
{
	unsigned int i;
	struct ipt_table_info bootstrap = { 0, { 0, 0, 0 }, NULL, { } };

#ifdef DEBUG_IP_FIREWALL_LOCKING
	iptd_wlocks = iptd_rlocks = 0;
#endif

	ipt_match = &ipt_standard_match;
	ipt_target = &ipt_standard_reg;

	/* Ensure counter counts match. */
	ipt_table_info = &bootstrap;

	replace_table(&initial_table.repl, NULL);

	duprintf("Slot[0][0]: packets = %u|%u bytes = %u|%u\n",
		 (unsigned int)(ipt_table_info->slots[0][0].counters.pcnt>>32),
		 (unsigned int)ipt_table_info->slots[0][0].counters.pcnt,
		 (unsigned int)(ipt_table_info->slots[0][0].counters.bcnt>>32),
		 (unsigned int)ipt_table_info->slots[0][0].counters.bcnt);

	/* Register setsockopt */
	if (nf_register_sockopt(&ipt_sockopts) < 0) {
		duprintf("Unable to register sockopts.\n");
		return 0;
	}

	/* Register hooks */
	for (i = 0; i < sizeof(ipt_ops)/sizeof(struct nf_hook_ops); i++) {
		if (nf_register_hook(&ipt_ops[i]) < 0) {
			duprintf("Unable to register iptable on hook %u.\n",
				 i+NF_IP_LOCAL_IN);
			return 0;
		}
	}

#if defined(DEBUG_IP_FIREWALL) || defined(DEBUG_IP_FIREWALL_USER)
	printk("Firewall tables: ``So... how is this better than KDE?''.\n");
#endif
	return 1;
}

#ifdef MODULE
/* Default to no forward for security reasons. */
static int forward = NF_DROP;
MODULE_PARM(forward, "i");

int
init_module(void)
{
	if (forward < 0 || forward > NF_MAX_VERDICT) {
		printk("forward=0/1\n");
		return -EINVAL;
	}

	initial_table.entries[NF_IP_FORWARD - 2].targinfo.standard.verdict
		= -forward - 1;
	if (!init())
		return -EINVAL;

	return 0;
}

void
cleanup_module(void)
{
	unsigned long flags;
	unsigned int i;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);

	for (i = 0; i < sizeof(ipt_ops)/sizeof(struct nf_hook_ops); i++)
		nf_unregister_hook(&ipt_ops[i]);
	nf_unregister_sockopt(&ipt_sockopts);
	kfree(ipt_table_info);

	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
}
#else /* !MODULE */
__initfunc(void ip_fw_init(void))
{
	if (!init())
		panic("Failed to initialize iptables\n");
}

#endif
